#include "stdafx.h"
#include "MovieTransmiter.h"
#include "AutoLocker.h"
#include "PacketHandler.h"
#include "UserPacketHandler.h"
#include "Server.h"
#include "Util.h"
#include "Log.h"

MovieTransmiter::MovieTransmiter()
{

}

MovieTransmiter::~MovieTransmiter()
{

}

void
MovieTransmiter::Init(Client *client)
{
  this->m_client = client;

}

bool
MovieTransmiter::InitializeTransmission(const string &filePath, ERROR_TYPE *error)
{
  if (this->m_context.status == STAT_NONE)
  {
    Server *server = this->m_client->GetServer();
    DataBase &db = server->GetDataBase();
    DataBase::DATABASE_FAIL_REASON reason = DataBase::DFR_NONE;

    if (db.HasPermission(filePath, this->m_client->GetUserInfo().id, &reason))
    {
      string movieDir = server->GetEnv().movie_dir;
      DataBase::MOVIE_INFO info;
      string physicalPath;

      Util::AppendDirSeperater(&movieDir);

      if (db.GetMovieInfo(filePath, &info, &reason))
      {
        Util::BindPath(movieDir, filePath, &physicalPath);

        this->m_context.metaDataBlocks = info.metaDataBlocks;
        this->m_context.bitRate = info.bitRate;
        this->m_context.status = STAT_STARTED;
        this->m_context.fileSize = info.totalSize;
        this->m_context.file = CreateFileA(physicalPath.c_str(), GENERIC_READ,
          FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

        if (this->m_context.file != INVALID_HANDLE_VALUE)
        {
          return true;

        }
        else
        {
          *error = ET_NOT_EXIST_FILENAME;

          return false;

        }

      }
      else
      {
        if (reason != DataBase::DFR_NONE)
        {
          *error = DataBase::FailResonToErrorType(reason);

        }
        else
        {
          *error = ET_NOT_SUPPORTED_FORMAT;

        }

        return false;

      }

    }
    else
    {
      if (reason != DataBase::DFR_NONE)
      {
        *error = DataBase::FailResonToErrorType(reason);

      }
      else
      {
        *error = ET_NO_PERMISSION;

      }

      return false;

    }

  }
  else
  {
    *error = ET_TRANSMIT_ALREADY_STARTED;

    return false;

  }

}

bool
MovieTransmiter::StartTransmission(long long startOffset, int size, ERROR_CODE *error)
{
  if (this->m_context.status == STAT_NONE)
  {
    //̸    .
    error->type = ET_TRANSMIT_NOT_STARTED;

    return false;

  }
  else
  {
    //࿡ ̹ û ؽƮ Ѵ.
    if (this->m_context.startOffset == startOffset && this->m_context.totalSize == size)
    {
      error->type = ET_TRANSMIT_ALREADY_TRANSMITING;

      return false;

    }
    else
    {
      if (startOffset <= this->m_context.fileSize)
      {
        LARGE_INTEGER offset;

        offset.QuadPart = startOffset;

        if (SetFilePointerEx(this->m_context.file, offset, NULL, FILE_BEGIN))
        {
          this->m_context.startOffset = startOffset;
          this->m_context.totalSentBytes = 0;
          this->m_context.totalSize = size;
          this->m_context.status = STAT_TRANSMITTING;
          this->m_context.lastTime = timeGetTime();
          this->m_context.lastSentBytes = 0;
          
          if (Util::DeterminMovieMetaDataRange(startOffset, size, this->m_context.metaDataBlocks))
            this->m_context.fast = true;
          else
            this->m_context.fast = false;

#ifdef _DEBUG
          cout << "Start offset is updated " << startOffset << " size " << size << endl;

          this->m_context.seq = 0;
          this->m_context.prevSent = 0;
          this->m_context.accumTime = 0.0;
#endif // _DEBUG

          return true;

        }
        else
        {
          Log::GetInstance().Write("SetFilePointerEx failed : \"%s\"",
            Util::SystemErrorToString(GetLastError()).c_str());

          error->type = ET_CANNOT_START_TRANSMISSION;
          error->errorCode = GetLastError();

          this->ResetContext(STAT_STARTED);

          return false;

        }

      }
      else
      {
        error->type = ET_EXCEED_OFFSET;

        this->ResetContext(STAT_STARTED);

        return false;

      }

    }

  }

}

bool
MovieTransmiter::StopTransmission()
{
  if (this->m_context.status == STAT_TRANSMITTING || this->m_context.status == STAT_STARTED)
  {
    this->ResetContext(STAT_NONE);

    CloseHandle(this->m_context.file);

    this->m_context.file = INVALID_HANDLE_VALUE;
    this->m_context.fileSize = 0ll;
    this->m_context.bitRate = 0;

    return true;

  }
  else
  {
    return false;

  }

}

void
MovieTransmiter::ResetContext(STATUS status)
{
  this->m_context.status = status;
  this->m_context.startOffset = -1ll;
  this->m_context.totalSize = -1;
  this->m_context.totalSentBytes = 0;
  this->m_context.lastTime = 0;
  this->m_context.lastSentBytes = 0;

#ifdef _DEBUG
  this->m_context.seq = 0;
  this->m_context.prevSent = 0;
  this->m_context.accumTime = 0.0;
#endif // _DEBUG

}

void
MovieTransmiter::TransmitStream(DWORD curTime)
{
  char buf[MAX_STREAM_SIZE];
  DWORD read = 0;
  PacketHandler parent(this->m_client);
  UserPacketHandler handler(this->m_client, &parent);

  if (ReadFile(this->m_context.file, buf, MAX_STREAM_SIZE, &read, NULL))
  {
    long long blockOffset = this->m_context.totalSentBytes + this->m_context.startOffset;
    bool ended = false;

    if (read < MAX_STREAM_SIZE)
    {
      //    ̴.
      //end ó Ѵ.
      ended = true;

    }
    else
    {
      if (this->m_context.totalSentBytes + (int)read >= this->m_context.totalSize)
      {
        //ٺ. end ó Ѵ.
        ended = true;

        read = this->m_context.totalSize - this->m_context.totalSentBytes;

      }

    }

    if (handler.SendStream(this->m_context.fast, blockOffset, read, buf))
    {
      this->m_context.lastTime = curTime;
      this->m_context.lastSentBytes = read;

      this->m_context.totalSentBytes += read;

      if (ended)
      {
        handler.SendStreamEnd();

        this->ResetContext(STAT_STARTED);

#ifdef _DEBUG
        cout << "Stream end" << endl;
#endif // _DEBUG

      }

    }
    else
    {
      // .  ߴ Ѵ.
      this->StopTransmission();
      this->ResetContext(STAT_NONE);

    }

  }
  else
  {
    //Ͽ  Ѵ. ó
    parent.SendError(ET_CANNOT_READ_FILE, GetLastError());

    this->ResetContext(STAT_STARTED);

  }

}

void
MovieTransmiter::Transmit()
{
  // ŭ о Ѵ.
  if (this->m_context.status == STAT_TRANSMITTING)
  {
    if (this->m_client->IsSendable())
    {
      if (this->m_context.bitRate == 0 || this->m_context.fast)
      {
        this->TransmitStream(0);

#ifdef _DEBUG
        cout <<  this->m_context.seq << ". Next block offset " <<
          this->m_context.totalSentBytes + this->m_context.startOffset <<
          " Send bytes " << this->m_context.totalSentBytes << endl;

        this->m_context.seq++;
#endif // _DEBUG

      }
      else
      {
        //ƮƮ Ѵٸ  ӵ Ͽ
        // ӵ Ѿ ŵѴ.
        int multiplier = this->m_client->GetServer()->GetEnv().buffering_multiplier;
        double bitsToByte = (double)this->m_context.bitRate / 8.0 * multiplier;
        DWORD curTime = timeGetTime();
        double deltaTime = curTime - this->m_context.lastTime;

        if (deltaTime < 0)
        {
          this->m_context.lastTime = 0;
          deltaTime = curTime - this->m_context.lastTime;

        }

        double deltaToSend = bitsToByte * deltaTime / 1000.0;

        if (this->m_context.lastSentBytes > deltaToSend)
        {
          return;

        }
        else
        {
          this->TransmitStream(curTime);

#ifdef _DEBUG
          if (this->m_context.accumTime >= 1000.0)
          {
            cout <<  this->m_context.seq << ". Next block offset " <<
              this->m_context.totalSentBytes + this->m_context.startOffset <<
              " Send bytes " << this->m_context.totalSentBytes - this->m_context.prevSent <<
              ", Delta time " << this->m_context.accumTime << endl;

            this->m_context.seq++;

            this->m_context.prevSent = this->m_context.totalSentBytes;
            this->m_context.accumTime = deltaTime;

          }
          else
          {
            this->m_context.accumTime += deltaTime;

          }
#endif // _DEBUG

        }

      }

    }
    else
    {
#ifdef _DEBUG
      cout << "Transmission panding" << endl;
#endif // _DEBUG

    }

  }

}
